/*
usbpower.c by Jochen Arndt <joe127@t-online.de>

Portions of this code are from usbutils, authors
 Thomas Sailer, <t.sailer@alumni.ethz.ch>
 Johannes Erdfelt
 David Brownell
 Aurelien Jarno

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdio.h>
#include <unistd.h>
#include <usb.h>
#include <error.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>

#define CTRL_TIMEOUT		(5*1000)	/* milliseconds */
#define	HUB_STATUS_BYTELEN	3		/* max 3 bytes status = hub + 23 ports */
#define USB_PORT_FEAT_POWER	8

#define PID_FILE "/var/run/usbpower.pid"
#define SLEEP_TIME		100		// msec
#define DELAY_TIME		30000		// msec

static int nStatus = 0;

#if 0
// Get number of ports
int get_port_count(struct usb_dev_handle *fd)
{
	unsigned char buf [7 /* base descriptor */
			+ 2 /* bitmasks */ * HUB_STATUS_BYTELEN ];
	int ret;

	ret = usb_control_msg(fd,
			USB_ENDPOINT_IN | USB_TYPE_CLASS | USB_RECIP_DEVICE,
			USB_REQ_GET_DESCRIPTOR,
			0x29 << 8, 0,
			buf, sizeof buf, CTRL_TIMEOUT);
	if (ret < 9 /* at least one port's bitmasks */) {
		if (ret >= 0)
			fprintf(stderr,
				"incomplete hub descriptor, %d bytes\n",
				ret);
		/* Linux returns EHOSTUNREACH for suspended devices */
		else if (errno != EHOSTUNREACH)
			perror ("can't get hub descriptor");
		return -1;
	}
	return (int)buf[2];
}
#endif

int show_port_status(struct usb_dev_handle *fd, int nPort)
{
	int ret;
	unsigned char stat[4];

	ret = usb_control_msg(fd,
		USB_ENDPOINT_IN | USB_TYPE_CLASS | USB_RECIP_OTHER,
		USB_REQ_GET_STATUS,
		0, nPort,
		stat, sizeof stat,
		CTRL_TIMEOUT);
	if (ret < 0) {
		fprintf(stderr,
			"usbpower: cannot read port %d status, %s (%d)\n",
			nPort, strerror(errno), errno);
	}

	printf("USB Port %d: %02x%02x.%02x%02x", nPort,
		stat[3], stat [2],
		stat[1], stat [0]);
	/* CAPS are used to highlight "transient" states */
	printf("%s%s%s%s%s",
		(stat[2] & 0x10) ? " C_RESET" : "",
		(stat[2] & 0x08) ? " C_OC" : "",
		(stat[2] & 0x04) ? " C_SUSPEND" : "",
		(stat[2] & 0x02) ? " C_ENABLE" : "",
		(stat[2] & 0x01) ? " C_CONNECT" : "");
	printf("%s%s%s%s%s%s%s%s%s%s\n",
		(stat[1] & 0x10) ? " indicator" : "",
		(stat[1] & 0x08) ? " test" : "",
		(stat[1] & 0x04) ? " highspeed" : "",
		(stat[1] & 0x02) ? " lowspeed" : "",
		(stat[1] & 0x01) ? " power" : "",
		(stat[0] & 0x10) ? " RESET" : "",
		(stat[0] & 0x08) ? " oc" : "",
		(stat[0] & 0x04) ? " suspend" : "",
		(stat[0] & 0x02) ? " enable" : "",
		(stat[0] & 0x01) ? " connect" : "");
	return ret;
}

struct usb_device *get_usb_device(int busnum, int devnum)
{
	struct usb_bus *bus;
	struct usb_device *dev;
	
	for (bus = usb_busses; bus; bus = bus->next) {
        	for (dev = bus->devices; dev; dev = dev->next) {
			if (atoi(bus->dirname) == busnum && atoi(dev->filename) == devnum)
				return dev;
		}
        }
	fprintf(stderr, "usbpower: Device %d on bus %d not found\n", devnum, busnum);
        return NULL;
}

// nPorts: Bit mask for ports; Bit 0 = port 1, bit 1 = port 2, ...
//         Use 0x06 for ports 2 und 3 (rear side of UFS-910)
// nOn:    0 = power off, 1 = power on, otherwise show only port status 
int power_usb(int nPorts, int nOn)
{
	int nPort = 0;
	int nResult = 0;
	int nLoop;
	struct usb_device *dev;
	struct usb_dev_handle *udev;

	if (nPorts == 0)
		return 0;
	usb_init();
	usb_find_busses();
	usb_find_devices();

	// The USB Hub is device #2 on bus #1
	dev = get_usb_device(1, 2);
	if (NULL == dev)
		return 1;

	udev = usb_open(dev);
	if (!udev) {
		fprintf(stderr, "Couldn't open USB Hub device\n");
		return 1;
	}
	if (dev->descriptor.bDeviceClass != USB_CLASS_HUB) {
		fprintf(stderr, "Device is not a Hub\n");
		usb_close(udev);
		return 1;
	}
	nLoop = nPorts;
	// USB Hub of UFS-910 has 4 ports
//	int nMaxPort = get_port_count(udev);
//	while (nLoop && nPort < nMaxPort) {
	while (nLoop && nPort < 4) {
		++nPort;
		if (nLoop & 1) {
			if (nOn == 1) {
				printf("Activating power for USB port %d\n", nPort);
				nResult = usb_control_msg(udev,
				    USB_TYPE_CLASS | USB_RECIP_OTHER,
				    USB_REQ_SET_FEATURE,
				    USB_PORT_FEAT_POWER,
				    nPort,
				    NULL, 0,
				    CTRL_TIMEOUT);
			}
			else if (nOn == 0) {
				printf("De-activating power for USB port %d\n", nPort);
				nResult = usb_control_msg(udev,
				    USB_TYPE_CLASS | USB_RECIP_OTHER,
				    USB_REQ_CLEAR_FEATURE,
				    USB_PORT_FEAT_POWER, // feature
				    nPort,
				    NULL, 0,
				    CTRL_TIMEOUT);
			}
		}
		nLoop >>= 1;
	}
	// Wait for power to be enabled
	// Hub of UFS-910 requires 50 * 2 msec
	if (nOn == 1)
		usleep(50 * 2 * 1000);
	// Show (new) port states after wait time
	nLoop = nPorts;
	nPort = 0;
	while (nLoop && nPort < 4) {
		++nPort;
		if (nLoop & 1)
			show_port_status(udev, nPort);
		nLoop >>= 1;
	}
	usb_close(udev);
	
	return nResult;
}

/* Write PID to lock file. */
int create_pid_file(void)
{
	FILE * f_pid;
	
	if (NULL != (f_pid = fopen(PID_FILE, "wb"))) {
		fprintf(f_pid, "%d\n", getpid());
    		fclose(f_pid);
        }
//        else
//		log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "open PID file");
	return f_pid == NULL ? -1 : 0;
}
						
/* Read PID from lock file. */
pid_t read_pid_file(void)
{
	pid_t pid = 0;
	char sBuf[16];
	FILE *f_pid;
        	
	if (NULL != (f_pid = fopen(PID_FILE, "rb"))) {
    		if (fgets(sBuf, 16, f_pid))
        		pid = (pid_t)atoi(sBuf);
    		fclose(f_pid);
	}
	return pid;
}

/* Signal function
   Use only save functions inside this handler (see man 2 signal).
*/
static void signal_handler(int signum)
{
	nStatus = signum;
}
    
int main(int argc, char *argv[])
{
	int i, j;
	int nRet = 0;
	int nPorts = 0;
	int nAction = -1;
	int nKill = 0;
	int nPOffOnStart = 0;
	int nPOffOnTerm = 0;
	unsigned nDelay = DELAY_TIME;
	unsigned nTime = SLEEP_TIME;
	pid_t pid = 0;
	
	// options:
	// (none): Show power settings
	// -0: Power off now
	// -1: Power on now
	// -k: Kill running daemon
	// -s: Power down on start
	// -t: Power down on term
	// -pN: Ports (bit 0 = port1, bit 1 = port 2, ...)
	// -dN: Delay time in seconds (min 5)
	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-') {
			nRet = 1;
			fprintf(stderr, "%s: Command line error\n", argv[0]);
		}
		else {
			j = 1;
			if (argv[i][j] == '\0') {
				nRet = 1;
				fprintf(stderr, "%s: Missing option after '-'\n", argv[0]);
				j = -2;
			}
			while (j >= 0 && argv[i][j]) {
				switch (argv[i][j]) {
					case '0' : nAction = 0; break;
				        case '1' : nAction = 1; break;
					case 'k' : nKill = 1; break;
				        case 's' : nPOffOnStart = 1; break;
					case 't' : nPOffOnTerm = 1; break;
				        case 'p' :
				    	        if (argv[i][j+1])
						        nPorts = atoi(argv[i] + j + 1);
						else if (i + 1 < argc)
						        nPorts = atoi(argv[++i]);
						else {
							fprintf(stderr, "%s: Missing parameter for option -p\n", argv[0]);
							nRet = 1;
						}
						j = -2;
						break;
				        case 'd' :
				    	        if (argv[i][j+1])
						        nDelay = 1000 * atoi(argv[i] + j + 1);
						else if (i + 1 < argc)
						        nDelay = 1000 * atoi(argv[++i]);
						else {
							fprintf(stderr, "%s: Missing parameter for option -d\n", argv[0]);
							nRet = 1;
						}
						j = -2;
						break;
					default :
						fprintf(stderr, "%s: Unknown option %c\n", argv[0], argv[i][j]);
						nRet = 1;
					
				}
				++j;
			}
		}
	}
	if (nDelay < 5000) {
		fprintf(stderr, "%s: Delay time too short\n", argv[0]);
		nRet = 1;
	}
	if (nRet)
		return nRet;
	if (nPOffOnStart || nPOffOnTerm) {
		if (nAction != -1)
			fprintf(stderr, "%s: Ignoring power option when daemonizing\n", argv[0]);
		nAction = -1;
	}
	if (nAction == -1 && !nPOffOnStart && !nPOffOnTerm && !nKill && nPorts)
		nAction = 2;
	if (nAction != -1)
		nRet = power_usb(nPorts, nAction);

	// Read PID file to check if already running as daemon
	pid = read_pid_file();

	// Check for stale lock file
	if (pid && kill(pid, 0)) {
		if (errno == ESRCH) {
			fprintf(stderr, "%s: Found stale lock file with PID %d\n", argv[0], pid);
			unlink(PID_FILE);
	    		pid = 0;
		}
	}
	
	// Kill daemon and exit if kill option specified
	if (nKill) {
		if (pid) {
			if (kill(pid, SIGUSR1)) {
				perror("usbpower: kill daemon");
				nRet = 1;
			}
			else {
				printf("%s: Daemon with PID %d killed\n", argv[0], pid);
//				pid = 0;
			}
		}
		else
			printf("%s: No daemon running\n", argv[0]);
	}
	
	if ((nPOffOnStart || nPOffOnTerm) && nPorts == 0) {
		fprintf(stderr, "%s: No USB ports specified\n", argv[0]);
		nRet = 1;
	}
	if (nRet || (!nPOffOnStart && !nPOffOnTerm))
		return nRet;
	    
	// Check if daemon already running
	if (pid) {
		fprintf(stderr, "%s: Already running as daemon with PID %d\n", argv[0], pid);
		return 1;
	}

//	puts("usbpower: daemonizing");
	if (daemon(0, 1)) {
		fprintf(stderr, "%s: Daemonizing failed\n", argv[0]);
		return 1;
	}

	// Write new PID to lock file after forking
	create_pid_file();

	// Install signal handler. Ignore errors (must kill manually).
	signal(SIGTERM, signal_handler);
	signal(SIGUSR1, signal_handler);
//	if (signal(SIGTERM, signal_handler) == SIG_ERR)
//		log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "Signal handler installation");

	// Power down defined USB ports upon start
	if (nPOffOnStart && nPorts)
		power_usb(nPorts, 0);

	while (!nStatus && nPorts) {
		// power up USB ports if started upon booting and time has passed
		usleep(SLEEP_TIME * 1000);
		if (nPOffOnStart && nTime && nPorts) {
			if (nTime >= nDelay) {
//				power_usb(nPorts, 2);
				power_usb(nPorts, 1);
				nTime = 0;
				if (!nPOffOnTerm)
					nStatus = SIGTERM;
			}
			else
				nTime += SLEEP_TIME;
		}
	}
	// Power down USB ports when terminating and option specified
	if (nStatus == SIGTERM && nPOffOnTerm && nPorts)
		power_usb(nPorts, 0);
//	puts("usbpower: Terminating.");
	unlink(PID_FILE);
	return 0;
}
